--GDL

-- http://en.wikipedia.org/wiki/Youtube#Quality_and_codecs
local qDesc = {
	    [5] = '  224p FLV'
	,   [6] = '  270p FLV'
	,  [34] = '  360p FLV'
	,  [35] = '  480p FLV'
	,  [18] = '  360p MP4'
	,  [22] = '  720p MP4'
	,  [37] = '1080p MP4'
	,  [38] = '2304p MP4'
	,  [43] = '  360p WebM'
	,  [44] = '  480p WebM'
	,  [45] = '  720p WebM'
	,  [46] = '1080p WebM'
	,  [13] = '3GP 176x144'
	,  [17] = '3GP 176x144'
	,  [36] = '3GP MP4'

	,  [83] = '  240p MP4 3D'
	,  [82] = '  360p MP4 3D'
	,  [85] = '  520p MP4 3D'
	,  [84] = '  720p MP4 3D'
	, [100] = '  360p WebM 3D'
	, [101] = '  480p WebM 3D'
	, [102] = '  720p WebM 3D'

    , [138] = 'mp4 no sound'
    , [137] = 'mp4 no sound'
    , [136] = 'mp4 no sound'
    , [135] = 'mp4 no sound'
    , [134] = 'mp4 no sound'
    , [133] = 'mp4 no sound'
    , [160] = 'mp4 no sound'
    , [264] = 'mp4 no sound'

	, [139] = 'audio mp4   48kb'
	, [140] = 'audio mp4 128kb'
	, [141] = 'audio mp4 256kb'
}

local conf = tableCombine({interactive=false, skip3D=true }, Settings[script.name] or {})
conf.qualityList = conf.qualityList or {38, 37, 22, 45, 44, 35, 43, 34, 18, 5}

local function getVideo(data, origUrl)
	local fmt = string.match(origUrl, '[?&]fmt=(%d+)')
	if fmt then fmt = tonumber(fmt) end
	origUrl = 'http://picasaweb.google.com/lh/sredir?uname='..data.feed.author[1]["gphoto$user"]["$t"]..'&target=photo&id='..data.feed["gphoto$id"]["$t"]

	local urls, info = {}, {}
	local title = utf8ToAnsi(data.feed.title["$t"]):gsub('\151','-')
	local newName = filterName(title, true)

	for i=#data.feed["media$group"]["media$content"], 2, -1 do
		local v = data.feed["media$group"]["media$content"][i]
		local itag = tonumber(v.url:match('[&?]itag=(%d+)'))

		if not ( conf.skip3D and ((qDesc[itag] or ''):find(' 3D$')) ) then
			if usd.progstatus() == 0 then return debugLog('USD stopped') end
			local sDesc = qDesc[itag] or v.type or '<unknown>'
			if conf.interactive and not fmt then
				debugSaveInc(logpref..'_video.html', usd.head(v.url), ('%s\r\n%s\r\n%s'):format(origUrl, sDesc, v.url) )
				urls[itag] = usd.getlocation()
			else
				urls[itag] = v.url
			end
			local size = usd.getheaders():match('HTTP/%d%.%d 200 OK') and usd.getheaders():match('Content%-Length: (%d+)')
			size = size and string.format('%0.2f', size / 1024 / 1024) or ' ??????'

			if conf.interactive and not fmt then
				table.insert(info
					, string.format(
						'%s: %s MB   ~%s   %s'
						, pad(itag,3, '  ')
						, pad(size, 7, '  ')
						, pad(v.width..'x'..v.height, 9, '  ')
						, sDesc
					)
				)
			end
		end
	end

	if fmt and urls[fmt] then
		debugSaveInc(logpref..'_video.html', usd.head(urls[fmt]), ('%s\r\n%s\r\n%s'):format(origUrl, fmt, qDesc[fmt] or '<unknown>') )
		dirUrl = usd.getlocation() .. '&title=' .. escapeUC(newName)
		dirUrl = addNewName(dirUrl, newName)
		usd.setretresult( dirUrl, 0, getResumeFlag() )
		return
	end

	if conf.interactive then
		local duration = math.floor(data.feed["gphoto$originalvideo"].duration / 60)..':'..pad(data.feed["gphoto$originalvideo"].duration % 60, 2)
		local dlg = {
			{duration, title}
			, {'__________________________'}
			, {'Select quality:'}
			, {{type='select', name='quality', info }}
		}
		dlg = Dialog.new(dlg, script.name)
		if usd.progstatus() == 0 then
			debugLog('USD stopped')
			return
		end
		local q = dlg:show()
		if q == '' then debugLog('Esc') return end		-- esc
		q = q:match('quality=([%d]+)') + 1	--     0
		q = tonumber(info[tonumber(q)]:match('^%s*(%d+)'))
		fmt = q
		usd.modquery(origUrl .. '&fmt='..fmt)
		dirUrl = urls[q] .. '&title=' .. escapeUC(newName)
		dirUrl = addNewName(dirUrl, newName)
		usd.setretresult( dirUrl, 0, 0 )
		return
	end

	if table.maxn(urls) == 0 then
		usd.setretresult('%Download link not found!')
		return
	end
	for i,v in ipairs(conf.qualityList) do
		if urls[v] then
			dirUrl = urls[v]
			fmt = v
			debugLog('GDL: url index: '..v)
			break
		end
	end
	if dirUrl == nil then
		fmt = table.maxn(urls)
		dirUrl = urls[fmt]
		debugLog('GDL: url index not found; selected max number: '..fmt)
	end

	usd.modquery(origUrl .. '&fmt='..fmt)
	debugSaveInc(logpref..'_video.html', usd.head(dirUrl), origUrl..'\r\n'..fmt)
	dirUrl = usd.getlocation() .. '&title=' .. escapeUC(newName)
	dirUrl = addNewName(dirUrl, newName)
	return usd.setretresult(dirUrl, 0, 0)

end


function GDL(origUrl) 	-- GetDirectLink
	logpref = script.name .. '_GDL_v03'
	usd.setstatus(script.name)
	local dirUrl, page, badpage, form, uri

	script.tmp = {}

	uri = URI.parse(origUrl)
	origUrl = uri()

	local data, user, album, photo, gupi
	local authkey = (uri.query or ''):match('authkey=[^&]+')
	authkey = authkey and '&'..authkey or ''

	if uri.host:find('%.ggpht%.com$')
		or uri.host:find('%.googleusercontent%.com$')
		or uri.host:find('%.bp%.blogspot%.com$')
		then
		--local tmp1, tmp2 = origUrl:match('^([^/]+//[^/]+/[^/]+/[^/]+/[^/]+/[^/]+)(.*)$')
		--tmp2='/d/'..filterName(tmp2:match('/([^/]+)$'))
		local tmp1, tmp2 = uri.path:match('^(/[^/]+/[^/]+/[^/]+/[^/]+)(.*)$')
		tmp2='/d/'..filterName(tmp2:match('/([^/]+)$'))
		--origUrl = tmp1..tmp2
		uri.path = tmp1..tmp2
		origUrl = uri()
		usd.modquery(origUrl)
		usd.setretresult(origUrl,0,1)
		return
	end

	local function checkUser()
		page = usd.get('http://picasaweb.google.com/data/entry/tiny/user/'..user..'?alt=jsonm&prettyprint=true'..authkey)
		data = parseJsObject(page, true, true)
		debugSaveInc(logpref..'.html', page, 'checkUser\r\n\r\n'..serialize(data or {}, nil, true))
		if not data then
			page = usd.get(origUrl)
			debugSaveInc(logpref..'.html', page, 'checkUser')
			user = page:match('//plus.google.com/photos/(%d+)/albums/')
		end
	end


--http://picasaweb.google.com/lh/sredir?uname=123&target=ALBUM&id=456
	if uri.path=='/lh/sredir' then
		local target = uri.query:match('target=(%w+)'):lower()
		local id = ('&'..(uri.query or '')):match('[&?]id=(%w+)')
		if target == 'photo' then photo=id end
		if target == 'album' then album=id end
		user = uri.query:match('uname=(%w+)')
--http://picasaweb.google.com/lh/photo/ggggggggggggggggggggggggggggggggggggggggggg?feat=embedwebsite
	elseif uri.path:find('^/lh/photo/') then
		page = usd.get(origUrl)
		debugSaveInc(logpref..'.html', page, '^/lh/photo/')
		user, photo = page:match('/user/(%w+)/photoid/(%d+)')
		gupi = page:match('&gupi=[^&"]+')
--http://picasaweb.google.com/user/album
--http://picasaweb.google.com/123/abc
--http://picasaweb.google.com/123/abc#789
	elseif uri.host:find('^picasaweb%.google%.') then
		user = uri.path:match('^/([%w.]+)')
		album = uri.path:match('^/%w+/(%w+)')
		photo = (uri.fragment or ''):match('^(%d+)$')
		if album and not (user and photo) then
			page = usd.get(origUrl)
			debugSaveInc(logpref..'.html', page, '^picasaweb%.google%.')
			user, album = page:match('/user/(%w+)/albumid/(%d+)')
		end
--[[
http://plus.google.com/photos/123/albums
http://plus.google.com/photos/123/albums/456
http://plus.google.com/photos/123/photo/789
  https://plus.google.com/photos/user/albums/456/789#photos/user/albums/456
]]
	elseif uri.host=='plus.google.com' then
		user = (uri.fragment or uri.path):match('^/?photos/([^/#?&]+)')
			or uri.path:match('^/(%d+)')
		album = (uri.fragment or uri.path):match('^/?photos/[^/#?&]+/albums/(%d+)')
		photo = (uri.fragment or uri.path):match('^/?photos/[^/#?&]+/photo/(%d+)')
			or (uri.fragment or uri.path):match('^/?photos/[^/#?&]+/albums/%d+/(%d+)')
		checkUser()
	end

	if not user then return usd.setretresult('%not supported') end

	debugLog(serialize({user=user,album=album,photo=photo,gupi=gupi,authkey=authkey}))
--do return end

	if photo then
-- --http://picasaweb.google.com/data/entry/tiny/user/109780686446922422512/photoid/5898081871963893474?alt=jsonm&prettyprint=true
		page = usd.get(('http://picasaweb.google.com/data/feed/tiny/user/%s/photoid/%s?alt=json&prettyprint=true&fields=title,author,gphoto:*,media:*%s%s'):format(user,photo,gupi or '',authkey))
		data = parseJsObject(page, true, true) -- noNumKeys, noNumVals
		debugSaveInc(logpref..'_photo.html', page, serialize(data or {}, nil, true))
		--if data.feed.originalVideo then return getVideo(data, 'http://picasaweb.google.com/lh/sredir?uname='..user..'&target=photo&id='..photo) end
		if data.feed["gphoto$videostatus"] then return getVideo(data, origUrl) end
		local username = data.feed.author[1].name["$t"]
		local albumname = data.feed["gphoto$albumInfo"] and data.feed["gphoto$albumInfo"][1].title
		albumname = albumname and '-'..albumname or ''
		local title = data.feed.title["$t"]
		local newName = filterName(utf8ToAnsi(username..albumname..'-'..title),true)
		dirUrl = data.feed["media$group"]["media$content"][1].url:gsub('(/[^/]+)$','/d%1')
		dirUrl = addNewName(dirUrl, newName )
		usd.setretresult(dirUrl,0,1)
		return
	end

	if album then
		if album:find('%D') then
			if album=='profile' then album = 'ProfilePhotos' end
			page = usd.get('http://picasaweb.google.com/'..user..'/ProfilePhotos')
			debugSaveInc(logpref..'_album.html', page)
            album = page:match('/albumid/(%d+)')
		end
		--page = usd.get(('http://picasaweb.google.com/data/feed/tiny/user/%s/photoid/%s?alt=json&prettyprint=true&fields=title,author,gphoto:*,media:*%s%s'):format(user,photo,gupi or '',authkey))
		page = usd.get(("http://picasaweb.google.com/data/feed/tiny/user/%s/albumid/%s?alt=json&prettyprint=false&fields=title,gphoto:*,openSearch:totalResults,entry(title,gphoto:access, gphoto:albumid, gphoto:id)%s%s")
			:format(user, album, gupi or '',authkey))
		data = parseJsObject(page, true, true) -- noNumKeys, noNumVals
		debugSaveInc(logpref..'_album.html', page, serialize(data or {}, nil, true))
		local atitle = data.feed["gphoto$nickname"]['$t']..'-'..data.feed.title['$t']
		debugLog('totalResults     = '..data.feed["openSearch$totalResults"]['$t'])
		debugLog('#data.feed.entry = '..tostring(#data.feed.entry))
		for k, v in ipairs(data.feed.entry) do
			local url = ('http://plus.google.com/photos/%s/photo/%s?/%s%s%s'):format(
				user, v["gphoto$id"]["$t"], filterName(utf8ToAnsi(atitle..'-'..v.title["$t"]),true),  gupi or '',authkey )
			usd.addquery(url)
		end
		usd.setretresult('%done%')
		return
	end

	if user then
		page = usd.get('http://picasaweb.google.com/data/feed/tiny/user/'..user..'?alt=json&prettyprint=false&fields=entry(title,author/name,gphoto:id),openSearch:totalResults')
		data = parseJsObject(page, true, true) -- noNumKeys, noNumVals
		debugSaveInc(logpref..'_albums.html', page, serialize(data or {}, nil, true))
		debugLog('totalResults     = '..data.feed["openSearch$totalResults"]['$t'])
		debugLog('#data.feed.entry = '..#data.feed.entry)

		for k, v in ipairs(data.feed.entry) do
			local username = v.author[1].name["$t"] or 'noName'
			local albumname = v.title["$t"] or 'noTitle'
			local url = 'http://plus.google.com/photos/'..user..'/albums/'..v["gphoto$id"]["$t"]..'?/'..filterName(utf8ToAnsi(username..'-'..albumname),true)
			usd.addquery(url)
		end
		usd.setretresult('%done%')
		return
	end


	usd.setretresult('%plugindefect')
	return
end

function OS(origUrl, dirUrl, fullName)
	logpref = script.name .. '_OS'
	local newName = dirUrl:match('/([^/?]+)$')
	return usd.setretresult(fileRename(fullName, newName))
end

regDirSuccess( '//picasaweb.google.', GDL, OS)
regDirSuccess( '//plus.google.com/', GDL, OS)
regDirSuccess( '.ggpht.com/', GDL, OS)
regDirSuccess('.googleusercontent.com/', GDL, OS)
regDirSuccess( '.bp.blogspot.com/', GDL, OS)
