--GDL

--[=[

Settings['rutube_ru'] = {
	  interactive = false
	, bitrate = 'max'	-- 'max'  'min'
}

]=]

local function makeFragNames(base,segTable)
	local tmp = {}
	for s=1, #segTable do
		for f=1, segTable[s] do
			table.insert(tmp,base..'Seg'..s..'-Frag'..f)
		end
	end
	return tmp
end

function GDL(origUrl) 	-- GetDirectLink


	logpref = script.name .. '_GDL_v15'
	usd.setstatus(script.name)

	local conf = Settings[script.name] or { interactive = false, bitrate = 'max'}
	conf.bitrate = (conf.bitrate or 'max'):lower()

	script.tmp={}
	local page

	local fmt = string.match(origUrl, '#fmt=(%d+)')
	if fmt then
		fmt = tonumber(fmt)
		origUrl = origUrl:match('^([^#]+)')
	end

	local id = origUrl:match('/tracks/(%x+)')
		or origUrl:match('/trackinfo/(%x+)')
		or origUrl:match('^http://video.rutube.ru/(%d+)$')
		or origUrl:match('^http://rutube.ru/play/embed/(%d+)/?$')

	if not id then
		page = usd.get(origUrl)
		debugSaveInc(logpref..'.html', page)
		if usd.getheaders():find('HTTP/%d%.%d 404 Not Found') then
			usd.setretresult('%notfound%')
			return
		end
		id = page:match('"og:video" content="http://video.rutube.ru/(%d+)"')
			or page:match('<link rel="canonical" href="http://rutube.ru/video/(%x+)')
	end

	page = usd.get('http://rutube.ru/api/play/trackinfo/'..id..'/?referer=http%3A%2F%2Fvideo%2Erutube%2Eru%2F'..id..'&format=json')
	local data = parseJsObject(page) or {}
	debugSaveInc(logpref..'.html', page, serialize(data,nil,true) )

	if data.detail and data.detail.player_stub and data.detail.player_stub.title_en=="Video does not exist"
		or data.detail == "Not found"
		then
		usd.setretresult('%notfound%')
		return
	end

	data = {trackinfo = data}
	usd.setreferer('')
	page = usd.get('http://rutube.ru/video/embed/'..data.trackinfo.track_id)
	--data = parseJsObject(page:match('player_data%s*=%s*{.-</script>') or '') or {}
	data.options = page:match('<div[^<]-id="options"[^<]+</div>')
	data.options = data.options:match('data%-value="([^"]+)"')
	data.options = entityDecode(data.options)
	data.options = parseJsObject(data.options)
	debugSaveInc(logpref..'.html', page, serialize(data or '<nil>',nil,true) )


	--local size = page:match('<size>(%d+)</size>')
--         flv
	--local title = utf8ToAnsi( page:match('<title><!%[CDATA%[(.-)%]%]></title>') )
	local title = utf8ToAnsi( data.trackinfo.title )
	script.tmp.newName = filterName(title, true) .. '.flv'

	--local lstInfo = getLstInfo()
	--script.tmp.lstInfo = lstInfo
	local tempDir = Settings['<misc>'].tempDir and replaceVars(Settings['<misc>'].tempDir):gsub('\\$','')..'\\' or getLstInfo().saveto

	local duration = data.trackinfo.duration / 1000 or 0
	local size = data.trackinfo.size or 0

	local balancer = data.options.video_balancer.default
	page = usd.get(balancer)
	debugSaveInc(logpref..'.html', page)

	local base = trim( page:match('<baseURL>(.-)</baseURL>') )

	local medias = {}
	local bitrates = {}
	local media
	for m in page:gmatch('<media .->') do
		local t = {}
		for k, v in m:gmatch('(%w+)="(.-)"') do
			k = k:lower()
			if k == 'url' then k = 'href' end
			t[k] = tonumber(v) or v:gsub('&amp;','&')
		end
		medias[t.bitrate] = t
		table.insert(bitrates, t.bitrate)
	end

	debugSaveInc(logpref..'.html', serialize(medias,nil,true),'medias',false)

	table.sort(bitrates)

	debugLog('bitrates = ' .. table.concat(bitrates, ', '))

	if conf.interactive and not fmt then
		local info = {}
		for i=#bitrates, 1, -1 do
			local b = bitrates[i]
			local m = medias[b]
			table.insert(info, ('%s MB  ~ %s x %s @ %s'):format(
	            pad( ('~%0.2f'):format(b * duration / 8 / 1024), 8, '  ')
	            , pad(m.width, 4, '  ')
	            , pad(m.height, 4, '  ')
	            , pad(m.bitrate, 5, '  ')
			))
		end

		local function s2hms(n) return ('%02d:%02d:%02d'):format(math.floor(n/60/60),math.floor((n/60)%60),n%60) end
		local dlg = {
			{s2hms(duration), title}
			, {'__________________________'}
			, {'~'..('%0.2f'):format(size / 1024 / 1024) .. 'MB'}
			, {'Select quality:'}
			, {{type='select', name='quality', info }}
		}
		dlg = Dialog.new(dlg, script.name)
		local q = dlg:show()
		if q == '' then return end		-- esc
		q = tonumber(q:match('quality=([%d]+)')) + 1	--     0
		fmt = tonumber(info[q]:match('@%s*(%d+)$'))
	end

-----------

	if not fmt then
		local b = tonumber(conf.bitrate)
		if b then
			fmt = b
		elseif conf.bitrate == 'max' then
			fmt = bitrates[#bitrates]
		else
			fmt = bitrates[1]
		end
	end

	media = medias[fmt]
	if not media then
		fmt, media = next(medias)
	end

	debugLog('selected = ' .. fmt)

	usd.modquery(origUrl .. '#fmt='..fmt)

	local manifest = trim(media.url or media.href)
	manifest = URI.combine(base, manifest)()
	local mediaBase = manifest:gsub('\.f4m$',''):gsub('\.f4m%?.-$','')



	page = usd.get(manifest)
	debugSaveInc(logpref..'.html', page)

	local bs = trim(page:match('<bootstrapInfo[^>]->(.-)</bootstrapInfo>'))
	bs = base64.dec(bs)
	local asrt = '%z%z..asrt.%z%z%z%z(....)'
	local tmp1, tmp2 = bs:find(asrt)
	tmp1 = struct.unpack('>I4',bs:match(asrt))	-- SegmentRunEntryCount
	debugLog('segments='..serialize(tmp1))
	tmp1 = '>'..(' I4I4'):rep(tmp1)
	tmp1 = {struct.unpack(tmp1,bs,tmp2+1)}
	table.remove(tmp1) -- pos
	local segTable = {}
	for i=1, #tmp1, 2 do
		if not segTable[tmp1[i]] then
			segTable[tmp1[i]]=tmp1[i+1]
		end
	end
	debugLog('segTable='..serialize(segTable))
	script.tmp.segTable = segTable

	script.tmp.fragFilebase = tempDir..mediaBase:match('/([^/]+)$')

	local function getSize(url) return tonumber(usd.head(url):match('Content%-Length: (%d+)')) end

	local parts = makeFragNames(mediaBase, segTable)
	local files = makeFragNames(script.tmp.fragFilebase, segTable)

	for i=1, #parts do
		usd.setstatus('part '..i..'/'..#parts)
		local fsize = fileSize(files[i])
		if not fsize or (fsize < getSize(parts[i])) then
			usd.get(parts[i]..'|'..files[i])
		end
	end

	usd.setretresult(manifest, 0, 0)

end

function OS(origUrl, dirUrl, fullName)
	logpref = script.name .. '_OS'
	debugLog(serialize(script.tmp))
	os.remove(fullName)	-- manifest

--https://github.com/K-S-V/Scripts/

	local AUDIO=0x08
	local VIDEO=0x09
	local flvHeader = 'FLV\1\5\0\0\0\9\0\0\0\0'
	local tagHeaderLen = 11;
	local prevTagSize  = 4;
	local initialAP    = true;
	local initialVP    = true;
	local noFrameSkip  = false;
	local pAudioTS     = -1;
	local pVideoTS     = -1;
	local pAudioTagPos = 0;
	local pAudioTagLen = 0;
	local pVideoTagPos = 0;
	local pVideoTagLen = 0;
	local totalTagLen

	local flv = io.open(script.tmp.fragFilebase..'.flv', 'w+b')
	flv:write(flvHeader)

	function ReadBox(frag, pos)
		local boxSize, boxType = struct.unpack('>I4 c4', frag, pos)
		if boxSize == 1 then
			boxSize = struct.unpack('>I8', frag, pos+8) - 16
			pos = pos + 16
		else
			boxSize = boxSize-8
			pos = pos+8
		end
		local box = frag:sub(pos,pos+boxSize-1)
		pos = pos+boxSize
		return pos, boxType, box
	end

	local files = makeFragNames(script.tmp.fragFilebase, script.tmp.segTable)

	debugLog('Joining...')
	for i=1, #files do
		local msg = 'part '..i..'/'..#files
		usd.setstatus(msg)
		debugLog(msg..' '..files[i])
		local frag = fileRead(files[i])
		local firstAP = true
		local firstVP = true
		local fragPos, boxPos = 1, 1
		local box, boxType

		while (boxType~='mdat' and fragPos<#frag) do
			fragPos, boxType, box = ReadBox(frag, fragPos)
		end

		if boxType ~= 'mdat' then
			usd.savetolog('Skipping fragment (no mdat): '..files[i])
		else
			while boxPos < #box do
				local packetType, packetSize, packetTS = struct.unpack('>I1 I3 I3', box, boxPos)
				local totalTagLen = tagHeaderLen + packetSize + prevTagSize
				if packetType==AUDIO then
					if packetTS >= pAudioTS then
						if noFrameSkip or initialAP or (not firstAP and (packetSize > 8)) then
							if (not noFrameSkip and ((packetTS > 0) and (packetTS == pAudioTS))) then
								if (totalTagLen <= pAudioTagLen) then
									--debugLog(("Skipping overwrite of audio packet packetTS=%d pAudioTS=%d packetSize=%d"):format(packetTS, pAudioTS, packetSize))
								else
									flv:seek('set',pAudioTagPos)
									if (pVideoTagPos > pAudioTagPos) then pVideoTagPos = pVideoTagPos - pAudioTagLen end
								end
							end
							pAudioTagPos = flv:seek()
							flv:write(box:sub(boxPos, boxPos+totalTagLen-1))
							--debugLog(("AUDIO packetTS=%d pAudioTS=%d packetSize=%d pAudioTagPos=%d  1"):format(packetTS, pAudioTS, packetSize, pAudioTagPos));
							pAudioTS     = packetTS;
							pAudioTagLen = totalTagLen;
						else
							--debugLog(("Skipping small sized audio packet AUDIO packetTS=%d pAudioTS=%d packetSize=%d"):format(packetTS, pAudioTS, packetSize));
							initialAP, firstAP = false, false
						end
					else
						--debugLog(("Skipping audio packet in fragment %s AUDIO packetTS=%d pAudioTS=%d packetSize=%d"):format(fragments[i], packetTS, pAudioTS, packetSize));
					end
				elseif packetType==VIDEO then
					if (packetTS >= pVideoTS) then
						if (noFrameSkip or initialVP or (not firstVP and (packetSize > 0))) then
							if (not noFrameSkip and ((packetTS > 0) and (packetTS == pVideoTS))) then
								if (totalTagLen <= pVideoTagLen) then
									--debugLog(("Skipping overwrite of video packet packetTS=%d pVideoTS=%d packetSize=%d"):format(packetTS, pVideoTS, packetSize))
								else
									flv:seek('set',pVideoTagPos)
									if (pAudioTagPos > pVideoTagPos) then pAudioTagPos = pAudioTagPos-pVideoTagLen end
								end
							end
							pVideoTagPos = flv:seek()
							flv:write(box:sub(boxPos, boxPos+totalTagLen-1))
							--debugLog(("VIDEO packetTS=%d pVideoTS=%d packetSize=%d pVideoTagPos=%d  1"):format(packetTS, pVideoTS, packetSize, pVideoTagPos));
							pVideoTS = packetTS;
							pVideoTagLen = totalTagLen;
						else
							--debugLog(("Skipping small sized audio packet VIDEO packetTS=%d pVideoTS=%d packetSize=%d"):format(packetTS, pVideoTS, packetSize));
							initialVP, firstVP = false, false
						end
					else
						--debugLog(("Skipping video packet in fragment %s VIDEO packetTS=%d pVideoTS=%d packetSize=%d"):format(fragments[i], packetTS, pVideoTS, packetSize));
					end
				else
					--debugLog("Skipping packetType="..packetType);
				end
				boxPos = boxPos + totalTagLen;
			end
		end
		box, frag = nil
	end

	flv:close()

	debugLog('Deleting parts...')
	for i=1, #files do
		os.remove(files[i])
	end

	local _, err = os.rename(script.tmp.fragFilebase..'.flv', fullName)
	if err then error(err) end

	return usd.setretresult(fileRename(fullName, script.tmp.newName))
end

--regDirSuccess( '//rutube.ru/track', GDL, OS)
--regDirSuccess( '//video.rutube.ru/', GDL, OS)
--regDirSuccess( '//rutube.ru/video/', GDL, OS)
regDirSuccess( {'//rutube.ru/', '.rutube.ru/'}, GDL, OS)
