-- GDL

local version = 28
local logpref

-- http://www.genyoutube.net/formats-resolution-youtube-videos.html
local qDesc = {
	    [5] = '  224p FLV'
	,   [6] = '  270p FLV'
	,  [34] = '  360p FLV'
	,  [35] = '  480p FLV'
	,  [18] = '  360p MP4'
	,  [22] = '  720p MP4'
	,  [37] = '1080p MP4'
	,  [38] = '2304p MP4'
	,  [43] = '  360p WebM'
	,  [44] = '  480p WebM'
	,  [45] = '  720p WebM'
	,  [46] = '1080p WebM'
	,  [13] = '3GP 176x144'
	,  [17] = '3GP 176x144'
	,  [36] = '3GP MP4'

	,  [83] = '  240p MP4 3D'
	,  [82] = '  360p MP4 3D'
	,  [85] = '  520p MP4 3D'
	,  [84] = '  720p MP4 3D'
	, [100] = '  360p WebM 3D'
	, [101] = '  480p WebM 3D'
	, [102] = '  720p WebM 3D'

    , [138] = 'no sound mp4'
    , [266] = 'no sound mp4'
    , [137] = 'no sound mp4'
    , [136] = 'no sound mp4'
    , [135] = 'no sound mp4'
    , [134] = 'no sound mp4'
    , [133] = 'no sound mp4'
    , [160] = 'no sound mp4'
    , [264] = 'no sound mp4'

    , [272] = 'no sound webm'
    , [313] = 'no sound webm'
    , [271] = 'no sound webm'
    , [248] = 'no sound webm'
    , [247] = 'no sound webm'
    , [244] = 'no sound webm'
    , [243] = 'no sound webm'
    , [242] = 'no sound webm'

	, [139] = 'audio mp4   48kb'
	, [140] = 'audio mp4 128kb'
	, [141] = 'audio mp4 256kb'

	, [171] = 'audio webm'
}

local Exts = {}
local slowFmt = {[5]=5}
local function makeExts(tbl, ext, ...) for k,v in ipairs(arg) do tbl[v] = ext end end
local function makeExtsAudio(tbl, ext, ...) for k,v in ipairs(arg) do tbl[v] = ext; slowFmt[v]=v end end
makeExts(Exts, '.flv', 5,6,34,35)
makeExts(Exts, '.mp4', 18,22,37,38,82,83,84,85)
makeExts(Exts, '.m4v', 133,134,135,136,137,138,160,264,266)	-- video
makeExts(Exts, '.webm', 43,44,45,46,100,101,102,242,243,244,247,248,271,272,313)
makeExts(Exts, '.3gp', 13,17,36)
makeExtsAudio(Exts, '.webm', 171,172,249,250,251)
makeExtsAudio(Exts, '.m4a', 139,140,141)

local conf = tableCombine({
		  interactive = false
		, useFallbackHost = true
		, skip3D = true
		, adaptiveFmts = false
		, nameTemplate = '%author%-%title%'
		, apiKey = 'AIzaSyBIx32m9qxxKJ7Ky9fRMGEmCpv8s7toPUI' -- github.com/adeng21/Cuptastic
	}, Settings[script.name] or {})
conf.qualityList = conf.qualityList or {38, 37, 22, 45, 44, 35, 43, 34, 18, 5}
conf.subtitles = tableCombine({translateSrc='en'}, conf.subtitles or {})
local sconf = conf.subtitles
sconf.enable = sconf.xml or sconf.srt or sconf.sbv

local function parse(str, spaces)
	local t = {}
	for k, v in string.gmatch(str, '([^=&]+)=([^&]*)') do
		if spaces then v = v:gsub('%+', ' ') end
		t[k] = tonumber(v) or unescape(v)
	end
	return t
end

local function parseDataCommas(data)
	for _, n in ipairs({'url_encoded_fmt_stream_map', 'adaptive_fmts'}) do
		if data[n] then
			local tmp = {}
			for t in data[n]:gmatch('[^,]+') do
				table.insert(tmp, parse(t, true))
			end
			data[n] = tmp
		end
	end
end

local function parseData(str)
	local data = parse(str, true)
	if data.stream then data.stream = parse(data.stream) end
	parseDataCommas(data)
	return data
end


Storage[script.name] = Storage[script.name] or {}
local st = Storage[script.name]

function getStsDecoder(url, force)
	if not force and st.sd then return st.sts, st.sd end
	local page = usd.get(url)
	debugSaveInc(logpref..'.html', page, 'getStsDecoder1')
	local sts = page:match('"sts":(%d+)')
	sts = tonumber(page:match('"sts":(%d+)'))
	local h5 = page:match('"js":"([^"]-/html5player[%w-]-%.js)"')
	if not h5 then
		debugLog('html5player.js not found')
		st.sts, st.sd = 16667, "r(a,20);d(a,1);r(a,72);s(a,8);r(a,47);s(a,5);d(a,2);s(a,30);s(a,66)"
		return st.sts, st.sd
	end
	h5 = unescapeJS(h5):gsub('^.-(//.+)$','http:%1')
	page = usd.get(h5)
	debugSaveInc(logpref..'.html', page)
	local jsd = {page:match(';var %S+=({%w+:function%(a.-}});(function %S+%(a%){a=a%.split%(""%);(.-);return a%.join%(""%)};)')}
	local jsdf = parseJsObject(jsd[1])
	if not jsdf then
		debugLog('getStsDecoder() error')
		st.sts, st.sd = 16667, "r(a,20);d(a,1);r(a,72);s(a,8);r(a,47);s(a,5);d(a,2);s(a,30);s(a,66)"
		return st.sts, st.sd
	end
	for k, v in pairs(jsdf) do jsdf[k] = v:find('reverse') and 'r' or v:find('splice') and 'd' or v:find('var c=a%[0%]') and 's' or 'error1' end
	st.sd = jsd[3]:gsub('%w+%.(%w+)', function(f)return jsdf[f] or 'error2' end)
	st.sts = sts
	debugLog('st='..serialize(st,nil,nil,true))
	return st.sts, st.sd
end

function check403(data)
	if data.use_cipher_signature=='True' and parseHeaders(usd.getheaders()).CODE==403 then
		getStsDecoder('http://www.youtube.com/watch?v='..data.video_id, true)
		local err = 'signature decoding error?'
		usd.setretresult('%'..err)
		--error(err)
		return true
	end
end


local function decodeSignature(si)
	si = {si:byte(1,#si)}
	local d={r=function() local tmp={} for i=#si,1,-1 do tmp[#tmp+1]=si[i]end si=tmp end, s=function(n)si[1],si[n+1]=si[n+1],si[1]end, d=function(n)si={select(n+1, unpack(si))}end}
	for f, n in st.sd:gmatch('(%a)%(a,?(%d*)%)') do	d[f](tonumber(n)) end
	return string.char(unpack(si))
end


local function blocked(data)
	return data.errorcode
		and data.errorcode == 150
		and data.reason
		and (
			data.reason:find('^Embedding disabled')
			or data.reason:find('restricted from playback on certain sites')
			or data.reason:find('blocked it from display on this website')
			or data.reason:find('Playback on other websites has been disabled')
		)
end

local countryBlocked = function(data) return data.status=="fail" and data.errorcode==150 and data.reason:find('country') end

local hosts = {}

local function addVideo(id,usePadding,seconds,name)
	usd.addquery(('http://youtu.be/%s#%s%s_%s'):format(
		id
		, usePadding~=false and pad2(id, ('W'):rep(11)) or ''
		, seconds and s2hms(seconds) or ''
		, filterName(utf8ToAnsi(name),true))
	)
end

-- UC[%w_-]+
local function getChannel(channel)
	local total, added, private = 0, 0, 0
	usd.get('http://0.0.0.0/headers')
	local baseUrl = ('https://www.googleapis.com/youtube/v3/search?part=snippet&channelId=%s&maxResults=50&safeSearch=none&order=date&key=%s'):format(channel, conf.apiKey)
	local npt=''
	for i=1, 50 do
		local page = usd.get(baseUrl..npt)
		local data = parseJsObject(page)
		debugSaveInc(logpref..'.html', baseUrl..npt..'\r\n\r\n'..serialize(data or '<nil>',nil,true), baseUrl..npt, false)
		total = data.pageInfo.totalResults
		if total == 0 then break end
		usd.setstatus(added..' / '..total)
		for k, v in ipairs(data.items) do
			local title = v.snippet.title
			if title == "Private video" then
				private = private+1
			elseif v.id.playlistId then
				usd.addquery(('http://www.youtube.com/playlist?list=%s#%s%s'):format(v.id.playlistId, pad2(v.id.playlistId, ('W'):rep(24)) ,filterName(utf8ToAnsi(title),true)))
				added = added+1
			elseif v.id.videoId then
				addVideo(v.id.videoId, conf.usePadding, false, title)
				added = added+1
			else
				if not (v.id.channelId and v.id.channelId==channel) then debugLog('unknown: '.. serialize(v.id)) end
			end
		end
		npt = data.nextPageToken and '&pageToken='..data.nextPageToken
		if not npt then break end
	end
	debugLog(('total = %d, added = %d, private = %d'):format(total, added, private))
	return total, added, private
end

local function getPlaylist(id)
	local total, added, private = 0, 0, 0
	usd.get('http://0.0.0.0/headers')
	local baseUrl = ('https://www.googleapis.com/youtube/v3/playlistItems?part=snippet&playlistId=%s&maxResults=50&safeSearch=none&key=%s'):format(id, conf.apiKey)
	local npt=''
	for i=1, 50 do
		local page = usd.get(baseUrl..npt)
		local data = parseJsObject(page)
		debugSaveInc(logpref..'.html', baseUrl..npt..'\r\n\r\n'..serialize(data or '<nil>',nil,true), baseUrl..npt, false)
		if data.error then
			local err = data.error.code==404 and 'notfound%' or data.error.message
			usd.setretresult('%'..err)
			return 0, 0, 0
		end
		total = data.pageInfo.totalResults
		if total == 0 then break end
		usd.setstatus(added..' / '..total)
		for k, v in ipairs(data.items) do
			local title = v.snippet.title
			local rid = v.snippet.resourceId
			if title == "Private video" then
				private = private+1
			elseif rid.videoId then
				addVideo(rid.videoId, conf.usePadding, false, title)
				added = added+1
			else
				debugLog('unknown: '.. serialize(v.id))
			end
		end
		npt = data.nextPageToken and '&pageToken='..data.nextPageToken
		if not npt then break end
	end
	debugLog(('total = %d, added = %d, private = %d'):format(total, added, private))
	return total, added, private
end


function GDL(origUrl) -- GetDirectLink
	usd.setstatus(script.name)
	logpref = script.name .. '_GDL_v'..version
	local page, badpage, dirUrl
	script.tmp={}

	debugLog(serialize(conf,nil,nil,true))
	debugLog('st='..serialize(st,nil,nil,true))

	local qList = conf.qualityList

	usd.setcookie('PREF=hl=en;')

	uri = URI.parse(origUrl)

	--   
	local fmt = string.match(origUrl, '[?&]fmt=(%d+)')
	if fmt then fmt = tonumber(fmt) end

	origUrl = unescape(origUrl)


	local id = string.match(origUrl, '[/?&]v[/=]([%w_-]+)')
		or string.match(origUrl, '/embed/([%w_-]+)')
		or string.match(origUrl, 'http://youtu.be/([%w_-]+)')
	script.tmp.id = id

	if not id then
		local user = origUrl:match('/user/([%w_-]+)')
		local channel = origUrl:match('/channel/([%w_-]+)')
		if user then
			local tmp1 = ('https://www.googleapis.com/youtube/v3/channels?part=id&forUsername=%s&maxResults=50&safeSearch=none&key=%s'):format(user, conf.apiKey)
			local tmp2 = usd.get(tmp1)
			local tmp3 = parseJsObject(tmp2)
			debugSaveInc(logpref..'.html', tmp1..'\r\n\r\n'..serialize(tmp3 or '<nil>',nil,true), tmp1, false)
			if #tmp3.items~=1 then
				for k, v in ipairs(tmp3.items) do usd.addquery(('http://www.youtube.com/channel/%s#%s'):format(v.id, filterName(utf8ToAnsi(v.title),true)))	end
				return
			end
			channel = tmp3.items[1].id
			user = nil
		end
		if channel then
			local total, added, private = getChannel(channel)
			usd.setretresult(total==added+private and '%done%' or '%notfound%')
			return
		end

		local playlist = origUrl:find('/playlist') and origUrl:match('[?&]list=([^&#]+)')
		if playlist then
			local total, added, private = getPlaylist(playlist)
			usd.setretresult(total==added+private and '%done%' or '%notfound%')
			return
		end

		usd.setretresult('%ID not found')
		return
	end

	origUrl = 'http://www.youtube.com/watch?v=' .. id
	--usd.modquery(origUrl  .. (fmt and '&fmt='..fmt or ''))
	usd.setreferer(origUrl)

	local sts = getStsDecoder(origUrl)
	usd.setreferer(origUrl)

	local data

	for _, v in ipairs({'', 'http://www.google.com/', 'http://www.youtube.com/'}) do
		usd.setreferer(v)
		local url = 'http://www.youtube.com/get_video_info?sts='..sts..'&hl=en&el=adunit&ps=default&gl=US&video_id='..id..'&eurl='..v
		page = usd.get(url)
		data = parseData(page)
		debugSaveInc(logpref..'.html', page..'\r\n', serialize(data,nil,true))
		if not blocked(data) then break end
	end


--reason=Embedding+disabled+by+request%3Cbr%2F%3E%3Cu%3E%3Ca+href%3D%27http%3A%2F%2Fwww.youtube.com%2Fwatch%3Fv%3DPVbwIJ7oHnU%26feature%3Dplayer_embedded%27+target%3D%27_blank%27%3EWatch+on+YouTube%3C%2Fa%3E%3C%2Fu%3E&errorcode=150&storyboard_spec=http%3A%2F%2Fi1.ytimg.com%2Fsb%2FPVbwIJ7oHnU%2Fstoryboard3_L%24L%2F%24N.jpg%7C48%2327%23100%2310%2310%230%23default%23C0nRCH_zU8VHDt_DF3ufc5f13-g%7C62%2345%23443%2310%2310%2310000%23M%24M%23Seiqhw936iQkh5kRmfE46W9GsO8%7C124%2390%23443%235%235%2310000%23M%24M%238p2FYy2cpSAn8GkaycBSnBndpLE&status=fail&loudness=-31.888999939&errordetail=0
-- VO275GDVjvY: status="fail",errorcode=150,reason="      NBC Universal,         ."
-- JA4W1EwtyRM: status="fail",errorcode=100,reason="    ."

-- http://www.youtube.com/watch?v=F4P3j8dh2ro

	if blocked(data) then
		data.reason = data.reason:gsub('<[^>]+>',' ')
		usd.savetolog('Error: '..data.reason)
		usd.setreferer('')
		page = usd.get(origUrl..'&hl=en')
		debugSaveInc(logpref..'.html', page)
		data = page:match('"args": {.-}')
		if data then
			data = parseJsObject(data)
		else
			local tmp = page:match('<param name="flashvars" value="([^"]+)"') or ''	-- ie6
			tmp = entityDecode(tmp)
			data = parse(tmp)
		end

		if not ( data and next(data) ) then
			local err = page:match('<div id="verify%-actions">%s*<p>(.-)</p>')
				or page:match('Sign in to confirm your age')
				or 'plugindefect'
			err = utf8ToAnsi(err)
			return usd.setretresult('%'..err, 10)
		end

		parseDataCommas(data)
		debugSaveInc(logpref..'.html', serialize(data, nil, true), nil, false)

	elseif data.status=='fail' then
		local r = utf8ToAnsi(data.reason)
		r = r:gsub('<[^>]+>',' ')
		usd.savetolog(r)
		if data.errorcode==100
			or data.errorcode==2
			or r=='This video is not available.'
			or r:find('copyright')
			or conf.delCountryBlocked and countryBlocked(data)
			then
			usd.setretresult('%deleted%')
		elseif r == "Please sign in to view this video." then --errorcode = 150
			usd.setretresult('%notfound%')
		else
			usd.setretresult('%'..r, 10)
		end
		return
	end

	if data.requires_purchase==1 then return usd.setretresult('%notfound%')	end

	if ( page:find('rtmpe%%3Dyes') or page:find('rtmpe%%253Dyes') )
		and (page:find('conn=rtmpe') or page:find('conn%%3Drtmpe') )
		then
		usd.setretresult('%RTMPE is not supported')
		return
	end

	if ( not data.url_encoded_fmt_stream_map or #data.url_encoded_fmt_stream_map==0 ) and data.hlsvp then
		usd.setretresult('%hlsvp is not supported')
		return
	end

	if not data.url_encoded_fmt_stream_map then
		usd.setretresult('%Download link not found!')
		return
	end


	local title = conf.nameTemplate
	title = utf8ToAnsi( type(title)=='function' and title(data) or title:gsub('%%([%w_-]+)%%',data) ):gsub('\151','-')
	local newName = filterName( title, true )
	script.tmp.newName = newName
	title = title:gsub('&','&&')

	if data.ttsurl and sconf.enable then
		--local tmp = usd.get(data.ttsurl..'&type=list&tlangs=1'..(sconf.asrs and '&asrs=1' or ''))
		local tmp = usd.get(data.ttsurl..'&type=list&tlangs=1&asrs=1')
		local subtitles = {asrs={},tracks={},targets={}}

--<track id="1" name="English v0.1" lang_code="en" lang_original="English" lang_translated="English" lang_default="true" cantran="true"/>
		for t in tmp:gmatch('<track[^>]+>') do
			--local id = t:match('id="(%d+)"')
			local name = t:match('name="([^"]-)"')
			local code = t:match('lang_code="([^"]+)"')
			local asr = t:find('kind="asr"')
			if name and code then subtitles[asr and 'asrs' or 'tracks'][code] = entityDecode(name) end
		end

--<target id="44" lang_code="sw" lang_original="Kiswahili" lang_translated="Swahili"/>
		for t in tmp:gmatch('<target[^>]+>') do
			local code = t:match('lang_code="([^"]+)"')
			local lang = t:match('lang_translated="([^"]+)"')
			if lang and code then subtitles.targets[code] = entityDecode(lang) end
		end

		debugSaveInc(logpref..'.html', tmp..'\r\n', serialize(subtitles,nil,true))

		tmp = false
		for _,l in ipairs(sconf.langs) do
			if subtitles.tracks[l] then
				tmp = data.ttsurl..'&name='..escapeUC(subtitles.tracks[l])..'&lang='..l
				break
			end
		end

		if not tmp and sconf.asrs then
			for _,l in ipairs(sconf.langs) do
				if subtitles.asrs[l] then
					tmp = data.ttsurl..'&kind=asr&name='..escapeUC(subtitles.asrs[l])..'&lang='..l
					break
				end
			end
		end

		if not tmp and sconf.translate then
			local tfrom = subtitles.tracks[sconf.translateSrc] and sconf.translateSrc
				or sconf.asrs and subtitles.asrs[sconf.translateSrc] and sconf.translateSrc
				or next(subtitles.tracks)
				or sconf.asrs and next(subtitles.asrs)
			if tfrom then
				tmp = data.ttsurl..'&name='..(subtitles.tracks[tfrom] or subtitles.asrs[tfrom]..'&kind=asr')..'&lang='..tfrom..'&tlang='..sconf.translate
			end
		end

		local function getTime(n,s) return ('%02d:%02d:%02d%s%03d'):format(n/60/60, n/60%60, n%60, s, n%1*1000+.5) end
		local function xml2srt(xml)
			local tmp = {}
			for i,t in xml:gmatch('(<text[^>]+>)([^<]+)</text>') do
				local s, d = i:match('start="([%d.]+)"'), i:match('dur="([%d.]+)"')
				if s and d then table.insert(tmp, ('%s --> %s\n%s'):format(getTime(s,','), getTime(s+d,','), entityDecode(entityDecode(trim(t)))) ) end
			end
			return tableConcat(tmp,'\n','\n\n') .. '\n\n'
		end
		local function xml2sbv(xml)
			local tmp = {}
			for i,t in xml:gmatch('(<text[^>]+>)([^<]+)</text>') do
				local s, d = i:match('start="([%d.]+)"'), i:match('dur="([%d.]+)"')
				if s and d then table.insert(tmp, ('%s,%s\n%s\n'):format(getTime(s,'.'), getTime(s+d,'.'), entityDecode(entityDecode(trim(t)))) ) end
			end
			return table.concat(tmp,'\n')..'\n'
		end


		if tmp then
			script.tmp.sfile = usd.workdir..id
			local xml = usd.get(tmp)
			if xml:find('<transcript[^>]->%s*<text') then
				if sconf.xml then usd.savetofile(script.tmp.sfile..'.xml', xml) end
				if sconf.srt then usd.savetofile(script.tmp.sfile..'.srt', xml2srt(xml)) end
				if sconf.sbv then usd.savetofile(script.tmp.sfile..'.sbv', xml2sbv(xml)) end
			else
				debugSaveInc(logpref..'.html', xml, 'bad xml subtitles?')
			end
		end
	end




	local res = {}	-- 
	for k, v in string.gmatch(','..data.fmt_list, ',(%d+)/([%dx]+)') do
		res[tonumber(k)] = v
	end

	local urls = {} --         
	local info = {} --  
	local streams = {} --  [itag]=data

	function parseStreams(streamData, formats)
		formats = type(formats)=='table' and formats
		for _, stream in ipairs(streamData) do
			if usd.progstatus() == 0 then  debugLog('USD stopped');  return  end
			if stream.stream then stream.stream = parse(stream.stream) end
			stream.itag = tonumber(stream.itag)
		end

		for _, stream in ipairs(streamData) do
			if (not formats or tableContains(formats, stream.itag))	then

				local vuri = URI.parse(stream.url)
				vuri.scheme='http'
				vuri.q.signature = vuri.q.signature or stream.sig or stream.s and decodeSignature(stream.s) or nil
				vuri.host = type(conf.useFallbackHost)=='string' and conf.useFallbackHost
					or (conf.useFallbackHost and stream.fallback_host )
					or vuri.host
				if conf.rateBypass then vuri.q.ratebypass='yes' end
				vuri.query = tableConcat(vuri.q, '=','&')
				video = vuri()

				if not qDesc[stream.itag] then
					local tmp = {}
					if stream.quality_label then table.insert(tmp, (#stream.quality_label:match('^%d+')==3 and '  ' or '')..stream.quality_label)	end
					table.insert(tmp,stream.type)
					qDesc[stream.itag] = table.concat(tmp, ' ')
				end
				if not ( conf.skip3D and (qDesc[stream.itag]:find(' 3D$') or stream.stereo3d) ) then
					if usd.progstatus() == 0 then return debugLog('USD stopped') end
					streams[stream.itag] = stream
					local sDesc = qDesc[stream.itag]
					if conf.interactive and not fmt then
						--debugSaveInc(logpref..'.html', usd.head(video), ('%s\r\n%s\r\n%s\r\n%s'):format(
						debugSaveInc(logpref..'.html', urlHead(video, origUrl), ('%s\r\n%s\r\n%s\r\n%s'):format(
							origUrl
							, sDesc
							, video
							, serialize(stream, nil, true) )
						)
						if check403(data) then return 'restart' end
						urls[stream.itag] = usd.getlocation()
					else
						urls[stream.itag] = video
					end
					local size = usd.getheaders():match('HTTP/%d%.%d 200 OK') and usd.getheaders():match('Content%-Length: (%d+)')
					size = size and string.format('%0.2f', size / 1024 / 1024) or ' ??????'

					if conf.interactive and not fmt then
						local r = res[stream.itag] or stream.size or ''
						r = ('  '):rep(4-#(r:match('^%d+') or ''))..r..('  '):rep(4-#(r:match('%d+$') or ''))
						table.insert(info
							, string.format(
								'%s: %s MB   ~%s   %s'
								, pad(stream.itag,3, '  ')
								, pad(size, 7, '  ')
								, r --, pad(res[stream.itag] or stream.size, 9, '  ')
								, sDesc
							)
						)
					end
				end
			end
		end
	end -- parseStreams


	if parseStreams(data.url_encoded_fmt_stream_map, conf.qualityList) == 'restart' then return end
	if conf.adaptiveFmts and data.adaptive_fmts then parseStreams(data.adaptive_fmts, conf.adaptiveFmts) end

	local thumbUrl = data.iurlmaxres or data.iurlsd or data.iurlhq or data.iurl or data.iurlmq or data.thumbnail_url
	if conf.saveCover or (conf.interactive and iup) then
		usd.head(thumbUrl)
		if parseHeaders().CODE==200 then
			script.tmp.tfile = usd.workdir..id..'.jpg'
			usd.get(thumbUrl..'|'..script.tmp.tfile)
		end
	end

	local function getAudioFmt(tbl, atype) for k, v in ipairs(tbl) do if streams[v] and streams[v].type:find('^audio/'..atype) then return v end end end

	local function selectFmt(fmt, resume)
		usd.modquery(origUrl .. '&fmt='..fmt)
		debugSaveInc(logpref..'.html', urlHead(urls[fmt], origUrl), ('%s\r\n%s %s'):format(origUrl, fmt, qDesc[fmt] or '<unknown>') )
		if check403(data) then return end
		dirUrl = usd.getlocation()
		local name = newName .. (Exts[fmt] or '.flv')
		dirUrl = dirUrl .. '&title=' .. escapeUC(name)
		dirUrl = addNewName(dirUrl, name)

		local restart, info
		if resume==0 then
			info = getLstInfo()
			local fname = info.localfile and info.localfile .. '.pusd'
			if (fileSize(fname) or 0) > 0 then os.remove(fname) end	-- new download
		end

		if streams[fmt].type:find('^audio/mp4') and conf.dashAudioFix then script.tmp.dashAudioFix=true end

		if streams[fmt].fps and conf.dashMerge then -- dash video
			local t = streams[fmt].type:match('^video/(%w+)')
			script.tmp.ext = '.'..t	-- .mp4/.webm
			local audioFile = getLstInfo(true).saveto..newName..id..'.'..(t=='mp4' and 'm4a' or t)	-- .m4a/.webm
			script.tmp.audioFile = audioFile
			local fmta = getAudioFmt(conf.dashAudio or {140, 171}, t)
			debugLog(serialize({audio={fmta or '!!! no fmta!', streams[fmta].type},video={fmt, streams[fmt].type}}))
			debugSaveInc(logpref..'.html', urlHead(urls[fmta], origUrl), ('%s\r\n%s %s'):format(origUrl, fmta, qDesc[fmta] or '<unknown>') )
			if check403(data) then return end
			local h = parseHeaders()
			local audio = h.LOCATION
			if resume == 0 and h.SIZE~=(fileSize(audioFile) or 0)
				or h.SIZE < (fileSize(audioFile) or 0)
				then os.remove(audioFile)
			end
			if h.SIZE > (fileSize(audioFile) or 0) then
				if conf.chunkedDownload then
					chunkedDownload(audio, nil, {chunkSize=conf.chunkSize, referer=origUrl, saveTo=audioFile, status='audio' })
				else
					usd.setstatus(('Downloading audio: %0.2f MB'):format(h.SIZE/1024/1024))
					usd.setreferer(origUrl)
					usd.get(audio..'|'..audioFile)
				end
			else
				debugLog('audio already downloaded')
			end
		end

		usd.setreferer(origUrl)

		--if conf.chunkedDownload and slowFmt[fmt] then
		if conf.chunkedDownload and not dirUrl:find('ratebypass=yes') then
			debugSaveInc(logpref..'.html', urlHead(urls[fmt], origUrl), ('%s\r\n%s %s'):format(origUrl, fmt, qDesc[fmt] or '<unknown>') )
			dirUrl = usd.getlocation()
			restart, info = chunkedDownload(dirUrl, name, {chunkSize=conf.chunkSize, referer=origUrl, info=info, status='video'})
			if restart then return end
		end

		usd.setreferer(origUrl)
		debugSaveInc(logpref..'.html', usd.head(dirUrl))	--   
		dirUrl = addNewName(usd.getlocation(), name)
		usd.setreferer(origUrl)
		usd.setretresult( dirUrl, 0, 1 ) --  'Range:' : 416 Requested range not satisfiable
		return
	end


	if fmt and urls[fmt] then
		selectFmt(fmt, 1)
		return
	end

	if not next(urls) and data.url_encoded_fmt_stream_map[1].s then
		usd.setretresult('%Unsupported')
		return
	end


	if conf.interactive then
		local duration = s2hms(data.length_seconds)
		local q
		--if data.storyboard_spec then
		--	local sb = {data.storyboard_spec:match('^([^|]+).-#([^|#]+)$')}
		--	debugLog(sb[1]:gsub('$L',2):gsub('$N','M0')..'?sigh='..sb[2])
		--end

		if iup and script.tmp.tfile then
			local desktopSize = {plg.gettextextent()}
			local maxWidth, maxHeight = math.floor((desktopSize[3]-desktopSize[1])*0.99), math.floor((desktopSize[4]-desktopSize[2])*0.95)
			debugSaveInc(logpref..'.html', usd.head('http://i.ytimg.com/vi/'..id..'/1.jpg'))
			local useSmallThumbs = parseHeaders(usd.getheaders()).CODE==200
			if useSmallThumbs then
				for i=1, 3 do	usd.get('http://i.ytimg.com/vi/'..id..'/'..i..'.jpg|'..usd.workdir..'youtube'..i..'.jpg')	end
			end
			local thumbImage = iup.LoadImage(script.tmp.tfile)
			local thumbWidth, thumbHeight = tonumber(thumbImage.width), tonumber(thumbImage.height)
			local shiftPreview = tonumber(thumbHeight)<=480
			local thumb = iup.label{image = thumbImage, CX=0, CY=shiftPreview and 95 or 0}

			local dlgWidth = math.max(math.min(maxWidth, thumbWidth), 600)
			local dlgHeight = math.min(maxHeight, thumbHeight + (shiftPreview and 145 or 50) )
			debugLog(serialize({ desktop=desktopSize, max={maxWidth, maxHeight}, dlg={dlgWidth,dlgHeight}, thumb={thumbWidth, thumbHeight}, shiftPreview=shiftPreview }))
			local smallThumbsBox = {CX=0, CY=0, gap=5}
			local smallThumbs = {}
			if useSmallThumbs then for i=1, 3 do
				smallThumbs[i] = iup.LoadImage(usd.workdir..'youtube'..i..'.jpg')
				smallThumbsBox[i] = iup.label{image = smallThumbs[i]}
			end end
			local quality = iup.list(tableCombine(info,{value=1, dropdown='YES', VISIBLE_ITEMS=30}))
			local flagOk = false
			local bOK = iup.button{title='OK', size='50x',  action=function() flagOk = true; return iup.CLOSE end}
			local bCancel = iup.button{title='Cancel', action=function() return iup.CLOSE end}
			debugLog(('plg.usdhwnd=%x; plg.usdhwndLUD=%s'):format(plg.usdhwnd, tostring(plg.usdhwndLUD)))
			local dlg = iup.dialog{
				iup.vbox{
					iup.label{title=duration..'   '..title}
			        , iup.hbox{ quality, bOK, bCancel; gap=20 }
					--, iup.cbox{ iup.hbox(smallThumbsBox), thumb; SIZE=thumbWidth..'x'..thumbHeight} -- 3.14
					, iup.cbox{ thumb, iup.hbox(smallThumbsBox); SIZE=minMax(128*3+10, maxWidth, thumbWidth)..'x'..minMax(96*2, maxHeight-50, thumbHeight)} -- 2.7.1
					; gap=5
				}
				; title = plg.exename .. ' / '.. plg.plgname .. '.plg / '..script.name..'.lua'
				--, font='Courier New::16'
				, defaultenter=bOK, defaultesc=bCancel
			    , RASTERSIZE=dlgWidth..'x'..dlgHeight
			    --, NATIVEPARENT = plg.usdhwnd -- don't show dialog
			    --, NATIVEPARENT = plg.usdhwndLUD -- don't works?
			}
			--iup.SetAttribute(dlg, "NATIVEPARENT", plg.usdhwnd) -- don't show dialog
			--iup.SetAttribute(dlg, "NATIVEPARENT", plg.usdhwndLUD) -- don't show dialog

			dlg.MAXSIZE = math.floor((desktopSize[3]-desktopSize[1])*.95)..'x'..math.floor((desktopSize[4]-desktopSize[2])*.95)
			if Settings['<Dialog>'].sound then plg.playsound(Settings['<Dialog>'].sound, 0, Settings['<Dialog>'].soundflags) end
			if Settings['<Dialog>'].beep then playbeep(Settings['<Dialog>'].beep) end

			debugLog('iup.Popup(dlg)='..tostring(iup.Popup(dlg)))	-- 0=ok?
			q = quality.value
			if useSmallThumbs then for i=1, 3 do
				iup.Destroy(smallThumbs[i])
				if not usd.fdebug then os.remove(usd.workdir..'youtube'..i..'.jpg') end
			end end
			iup.Destroy(quality)
			iup.Destroy(bOK)
			iup.Destroy(bCancel)
			iup.Destroy(thumb)
			iup.Destroy(thumbImage)
			iup.Destroy(dlg)

			if not flagOk then
				debugLog('Cancel?')
				if not usd.fdebug then os.remove(script.tmp.tfile) end
				return
			end

		else
			local dlg = {
				{duration, title}
				, {'__________________________'}
				, {'Select quality:'}
				, {{type='select', name='quality', info }}
			}
			dlg = Dialog.new(dlg, script.name)
			if usd.progstatus() == 0 then
				debugLog('USD stopped')
				return
			end
			q = dlg:show()
			if q == '' then
				debugLog('Esc')
				if not usd.fdebug then os.remove(script.tmp.tfile) end
				return
			end
			q = q:match('quality=([%d]+)') + 1	--     0
		end -- IUP/SO

		q = tonumber(info[tonumber(q)]:match('^%s*(%d+)'))
		fmt = q

		selectFmt(fmt,0)

		return
	end -- interactive

	if table.maxn(urls) == 0 then
		usd.setretresult('%Download link not found!')
		return
	end
	for i,v in ipairs(qList) do
		if urls[v] then
			dirUrl = urls[v]
			fmt = v
			debugLog('GDL youtube: url index: '..v)
			debugSaveInc(logpref..'.html', urlHead(dirUrl, origUrl), origUrl..'\r\n'..fmt)
			if parseHeaders(usd.getheaders()).CODE==200 then	break	end
			if check403(data, 'auto1') then return end
		end
	end
	if dirUrl == nil then
		fmt = table.maxn(urls)
		dirUrl = urls[fmt]
		debugLog('GDL youtube: url index not found; selected max number: '..fmt)
		debugSaveInc(logpref..'.html', urlHead(dirUrl, origUrl), origUrl..'\r\n'..fmt)
		if check403(data, 'auto2') then return end
	end

	if not usd.fdebug then
		if not conf.saveCover and fileSize(script.tmp.tfile) then
			os.remove(script.tmp.tfile)
			script.tmp.tfile = nil
		end
	end


	selectFmt(fmt,0)

end

function OS(origUrl, dirUrl, fullName, headers)
	logpref = script.name .. '_OS_v'..version

	if script.tmp.dashAudioFix and conf.dashAudioFix then
		debugLog('Fixing...')
		local dest = fullName:gsub('\\([^\\]+)$','\\'..script.tmp.id..'.m4a')
		local cmd = ('"%s" "%s" "%s"'):format(conf.dashAudioFix, fullName, dest)
		plg.exec(cmd, true, SW_SHOWMINNOACTIVE)
		if (fileSize(dest) or 0)>0 then fullName = dest end
	end

	if script.tmp.audioFile and conf.dashMerge then
		debugLog('Merging...')
		local dest = fullName:gsub('\\([^\\]+)$','\\'..script.tmp.id..script.tmp.ext)
		local cmd = ('"%s" "%s" "%s" "%s"'):format(conf.dashMerge, fullName, script.tmp.audioFile, dest)
		plg.exec(cmd, true, SW_SHOWMINNOACTIVE)
		local sizes = {video=fileSize(dest) or -1, audio=fileSize(script.tmp.audioFile) or -1, result=fileSize(dest) or -1}
		debugLog(serialize(sizes))
		if sizes.result>1024 then fullName = dest end
	end

	local newName = fileRename(fullName, script.tmp.newName, getProperExt(fullName) or -1)

	if script.tmp.sfile then
		for _,v in ipairs({'.xml','.srt','.sbv'}) do
			local f=script.tmp.sfile..v
			if (fileSize(f) or 0)>24 then
				local _, err = os.rename(f, newName..v)
				if err then usd.savetolog(err) end
			end
		end
	end

	if script.tmp.tfile then
		local s = fileSize(script.tmp.tfile) or 0
		debugLog('thumb file size = '..tostring(s))
		local _, err = os.rename(script.tmp.tfile, newName..'.jpg')
		if err then usd.savetolog(err) end
	end

	usd.setretresult(newName)
end

function SU(url)
	local uri = URI.parse(url)
	if uri.host=='youtu.be' or uri.path == '/watch' then return false end
	return true
end

regDirSuccess('youtube.com/', GDL, OS, SU)
regDirSuccess('//youtu.be/', GDL, OS, SU)
